from typing import Iterable, Iterator, Sequence, TypeVar

from pipd import Function, Pipe

T = TypeVar("T")


class Unbatch(Function):
    def __call__(self, items: Iterable[Sequence[T]]) -> Iterator[T]:
        for b in items:
            for item in b:
                yield item


Pipe.add_fn(Unbatch)



from random import randint
from typing import Iterator, List, Optional, TypeVar

from pipd import Function, Pipe

T = TypeVar("T")


def pick(items: List[T], random: bool = False) -> T:
    idx = randint(0, len(items) - 1) if random else 0
    return items.pop(idx)


class Shuffle(Function):
    def __init__(self, size: int, start: Optional[int] = None):
        self.size = size
        self.start = start or size

    def __call__(self, items: Iterator[T]) -> Iterator[T]:
        buffer = []
        for item in items:
            buffer.append(item)
            if len(buffer) < self.size:
                try:
                    buffer.append(next(items))
                except StopIteration:
                    pass
            if len(buffer) >= self.start:
                yield pick(buffer, random=True)
        # Empty buffer at the end
        while len(buffer) > 0:
            yield pick(buffer)


Pipe.add_fn(Shuffle)



from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
from typing import Callable, Iterable, Iterator, TypeVar

from pipd import Function, Pipe, log_traceback_and_continue

T = TypeVar("T")
U = TypeVar("U")


class Side(Function):
    def __init__(
        self,
        fn: Callable[[T], U],
        num_workers: int = 0,
        mode: str = "multithread",
        handler: Callable = log_traceback_and_continue,
    ) -> None:
        assert mode in ["multithread", "multiprocess"]
        self.fn = fn
        self.num_workers = num_workers
        self.mode = mode
        self.handler = handler

    def __call__(self, items: Iterable[T]) -> Iterator[T]:
        if self.num_workers == 0:
            for item in items:
                try:
                    self.fn(item)
                except Exception as e:
                    self.handler(e)
                yield item
            return

        executors = dict(
            multithread=ThreadPoolExecutor, multiprocess=ProcessPoolExecutor
        )
        with executors[self.mode](max_workers=self.num_workers) as executor:
            for item in items:
                executor.submit(self.fn, item)
                yield item


Pipe.add_fn(Side)



from typing import Iterable, Iterator, TypeVar

from pipd import Function, Pipe

T = TypeVar("T")


class Tqdm(Function):
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __call__(self, items: Iterable[T]) -> Iterator[T]:
        try:
            from tqdm import tqdm as progressbar
        except ImportError:
            raise ImportError("tqdm is required to use tqdm")

        for item in progressbar(items, *self.args, **self.kwargs):
            yield item


Pipe.add_fn(Tqdm)



import asyncio
import glob
import os
import random
from typing import Iterable, Iterator, Optional, Sequence, TypeVar

from pipd import Function, Pipe

from .read_lines import read_lines
from .write_lines import write_lines

T = TypeVar("T")
U = TypeVar("U")


def watchdir(
    path: str, changes: Sequence[int] = [1]  # default to watchgod.Change.added == 1
) -> Iterator[str]:
    try:
        from watchgod import awatch
    except ImportError:
        raise ImportError("watchgod is required to use watch")
    loop = asyncio.get_event_loop()
    async_generator = awatch(path)
    while True:
        for change, filepath in loop.run_until_complete(async_generator.__anext__()):
            if change in changes:
                yield filepath


class ReadFiles(Function):
    def __init__(
        self,
        cache_filepath: Optional[str] = None,
        watch: bool = False,
        shuffle: bool = False,
    ) -> None:
        self.cache_filepath = cache_filepath
        self.watch = watch
        self.shuffle = shuffle

    def __call__(self, items: Iterable[str]) -> Iterator[str]:
        for filepath in items:
            files = []
            if self.cache_filepath is not None:
                if os.path.exists(self.cache_filepath):
                    files = list(read_lines(filepath=self.cache_filepath))
                else:
                    files = glob.glob(filepath)
                    list(write_lines(files, filepath=self.cache_filepath))
            else:
                files = glob.glob(filepath)

            if self.shuffle:
                random.shuffle(files)

            for file in files:
                yield file
            if self.watch:
                yield from watchdir(os.path.dirname(filepath))


Pipe.add_fn(ReadFiles)



from typing import Iterable, Iterator

from pipd import Function, Pipe


def write_lines(items: Iterable[str], filepath: str) -> Iterator[str]:
    with open(filepath, "w") as f:
        for item in items:
            f.write(f"{item}\n")
            yield item


class WriteLines(Function):
    def __init__(self, filepath: str) -> None:
        self.filepath = filepath

    def __call__(self, items: Iterable[str]) -> Iterator[str]:
        return write_lines(items, self.filepath)


Pipe.add_fn(WriteLines)



from pipd import Pipe


def test_pipe():
    pipe = Pipe(0, 1, 2, 3)
    assert list(pipe) == [0, 1, 2, 3]

    pipe = Pipe([0, 1, 2, 3])
    assert list(pipe) == [0, 1, 2, 3]

    pipe = Pipe(range(4))
    assert list(pipe) == [0, 1, 2, 3]

    class ClassWithGetItem:
        def __getitem__(self, key):
            return key

    pipe = Pipe(ClassWithGetItem())
    it = iter(pipe)
    assert next(it) == 0
    assert next(it) == 1
    assert next(it) == 2

    pipe = Pipe(0, 1, 2, 3)

    def gen():
        yield from pipe

    assert list(gen()) == [0, 1, 2, 3]


def test_metaclass():
    pipe = Pipe
    assert list(pipe(0, 1, 2, 3)) == [0, 1, 2, 3]

    pipe = Pipe.map(lambda x: x + 1)
    assert list(pipe(0, 1, 2, 3)) == [1, 2, 3, 4]


# def test_merge():
#     pipe0 = Pipe(0, 0, 0, 0, 0, 0)
#     pipe1 = Pipe(1, 1, 1, 1, 1, 1)
#     pipe = Pipe.merge(pipe0, pipe1, weights=[3, 1])
#     print(list(pipe))



from typing import Iterable, Iterator, List, TypeVar

from pipd import Function, Pipe

T = TypeVar("T")


class Batch(Function):
    def __init__(self, size: int, partial: bool = True) -> None:
        self.size = size
        self.partial = partial

    def __call__(self, items: Iterable[T]) -> Iterator[List[T]]:
        batch = []
        for item in items:
            batch.append(item)
            if len(batch) == self.size:
                yield batch
                batch = []
        if batch and self.partial:
            yield batch


Pipe.add_fn(Batch)



from concurrent.futures import (
    FIRST_COMPLETED,
    ProcessPoolExecutor,
    ThreadPoolExecutor,
    wait,
)
from typing import Callable, Iterable, Iterator, Optional, TypeVar

from pipd import Function, Pipe, log_traceback_and_continue

T = TypeVar("T")
U = TypeVar("U")


class Map(Function):
    def __init__(
        self,
        fn: Callable[[T], U],
        num_workers: int = 0,
        buffer: Optional[int] = None,
        mode: str = "multithread",
        handler: Callable = log_traceback_and_continue,
    ) -> None:
        assert mode in ["multithread", "multiprocess"]
        self.fn = fn
        self.num_workers = num_workers
        self.buffer = buffer
        self.mode = mode
        self.handler = handler

    def __call__(self, items: Iterable[T]) -> Iterator[U]:
        if self.num_workers == 0:
            for item in items:
                try:
                    yield self.fn(item)
                except Exception as e:
                    self.handler(e)
            return

        executors = dict(
            multithread=ThreadPoolExecutor, multiprocess=ProcessPoolExecutor
        )

        with executors[self.mode](max_workers=self.num_workers) as executor:
            futures = set()
            self.buffer = self.buffer or self.num_workers
            for item in items:
                futures.add(executor.submit(self.fn, item))
                if len(futures) == self.buffer:
                    done, futures = wait(futures, return_when=FIRST_COMPLETED)
                    for future in done:
                        try:
                            yield future.result()
                        except Exception as e:
                            self.handler(e)
            for future in futures:
                try:
                    yield future.result()
                except Exception as e:
                    self.handler(e)


Pipe.add_fn(Map)



from typing import Iterable, Iterator, TypeVar

from pipd import Function, Pipe

T = TypeVar("T")


class Limit(Function):
    def __init__(self, limit: int = 10**100) -> None:
        self.limit = limit

    def __call__(self, items: Iterable[T]) -> Iterator[T]:
        for count, item in enumerate(items):
            if count >= self.limit:
                return
            yield item


Pipe.add_fn(Limit)



from typing import Any, Dict, Iterable, Iterator, Sequence, Union

from pipd import Function, Pipe


class WriteCSV(Function):
    def __init__(self, filepath: str) -> None:
        self.filepath = filepath

    def __call__(
        self, items: Iterable[Union[Dict[str, Any], Sequence[Any]]]
    ) -> Iterator[Union[Dict[str, Any], Sequence[Any]]]:
        import csv

        with open(self.filepath, "w", newline="") as f:
            writer = csv.writer(f)
            for i, item in enumerate(items):
                if isinstance(item, dict):
                    if i == 0:  # Write headers only for the first dictionary item
                        writer.writerow(item.keys())
                    item = item.values()  # type: ignore
                writer.writerow(item)
                yield item


Pipe.add_fn(WriteCSV)



from typing import Dict, Iterable, Iterator, List, Sequence, Union

from pipd import Function, Pipe


class ReadCSV(Function):
    def __init__(self, header: Union[bool, Sequence[str]] = False) -> None:
        self.header = header

    def __call__(self, items: Iterator[str]) -> Iterable[Union[Dict[str, str], List[str]]]:  # type: ignore # noqa
        import csv

        for filepath in items:
            with open(filepath, "r") as f:
                if self.header:
                    fieldnames = (
                        self.header if isinstance(self.header, Sequence) else None
                    )
                    yield from csv.DictReader(f, fieldnames=fieldnames)
                else:
                    yield from csv.reader(f)


Pipe.add_fn(ReadCSV)



from typing import Callable, Iterable, Iterator, TypeVar

from pipd import Function, Pipe

from .map import Map

T = TypeVar("T")


class Filter(Function):
    def __init__(self, fn: Callable[[T], bool], *args, **kwargs) -> None:
        self.fn = lambda x: (x, fn(x))
        self.args = args
        self.kwargs = kwargs

    def __call__(self, items: Iterable[T]) -> Iterator[T]:
        for item, keep in Map(self.fn, *self.args, **self.kwargs)(items):  # type: ignore
            if keep:  # type: ignore
                yield item  # type: ignore


Pipe.add_fn(Filter)



import time
from typing import Iterable, Iterator, TypeVar

from pipd import Function, Pipe

from .side import Side

T = TypeVar("T")


class Sleep(Function):
    def __init__(self, seconds: float) -> None:
        self.seconds = seconds

    def __call__(self, items: Iterable[T]) -> Iterator[T]:
        return Side(lambda _: time.sleep(self.seconds))(items)  # type: ignore


Pipe.add_fn(Sleep)



from .batch import Batch
from .filter import Filter
from .filter_cached import FilterCached
from .limit import Limit
from .log import Log
from .map import Map
from .read_csv import ReadCSV
from .read_files import ReadFiles
from .read_lines import ReadLines
from .shuffle import Shuffle
from .side import Side
from .sleep import Sleep
from .tqdm import Tqdm
from .unbatch import Unbatch
from .write_csv import WriteCSV
from .write_lines import WriteLines



import select
from typing import Iterable, Iterator

from pipd import Function, Pipe


def read_lines(filepath: str, watch: bool = False) -> Iterator[str]:
    with open(filepath, "r") as file:
        if watch:
            while True:
                line = file.readline()
                if line:
                    yield line.strip()
                else:
                    select.select(
                        [file], [], []
                    )  # Wait until there is more data to read
        else:
            for line in file:
                yield line.strip()


class ReadLines(Function):
    def __init__(self, watch: bool = False) -> None:
        self.watch = watch

    def __call__(self, items: Iterable[str]) -> Iterator[str]:
        for filepath in items:
            yield from read_lines(filepath=filepath, watch=self.watch)


Pipe.add_fn(ReadLines)



from setuptools import find_packages, setup

setup(
    name="pipd",
    packages=find_packages(exclude=[]),
    version="0.1.6",
    description="Utility functions for python data pipelines.",
    long_description_content_type="text/markdown",
    author="ElevenLabs",
    url="https://github.com/elevenlabs/pipd",
    keywords=["data processing", "pipeline"],
    install_requires=[],
    classifiers=[],
)



from typing import Callable, Iterable, Iterator, TypeVar

from pipd import Function, Pipe

from .side import Side

T = TypeVar("T")


class Log(Function):
    def __init__(self, fn: Callable[[T], None] = print) -> None:
        self.fn = fn

    def __call__(self, items: Iterable[T]) -> Iterator[T]:
        return Side(self.fn)(items)


Pipe.add_fn(Log)



from pipd import Pipe


def double(x):
    return x * 2


def test_map():
    pipe = Pipe(range(5)).map(lambda x: x * 2)
    assert list(pipe) == [0, 2, 4, 6, 8]

    pipe = Pipe(range(5)).map(lambda x: x * 2, num_workers=2)
    assert sorted(pipe) == [0, 2, 4, 6, 8]

    # Multiprocess doesn't support lambdas or closures
    pipe = Pipe(range(5)).map(double, num_workers=2, mode="multiprocess")
    assert sorted(pipe) == [0, 2, 4, 6, 8]


def test_filter():
    pipe = Pipe(range(5)).filter(lambda x: x % 2 == 0)
    assert list(pipe) == [0, 2, 4]

    pipe = Pipe(range(5)).filter(lambda x: x % 2 == 0, num_workers=2)
    assert sorted(pipe) == [0, 2, 4]


def test_side():
    pipe = Pipe(range(5)).side(lambda x: x * 2)
    assert list(pipe) == [0, 1, 2, 3, 4]

    pipe = Pipe(range(5)).side(lambda x: x * 2, num_workers=2)
    assert list(pipe) == [0, 1, 2, 3, 4]


def test_batch():
    pipe = Pipe(range(5)).batch(2)
    assert list(pipe) == [[0, 1], [2, 3], [4]]

    pipe = Pipe(range(5)).batch(2, partial=False)
    assert list(pipe) == [[0, 1], [2, 3]]


def test_unbatch():
    pipe = Pipe([[0, 1], [2, 3], [4]]).unbatch()
    assert list(pipe) == [0, 1, 2, 3, 4]


def test_shuffle():
    pipe = Pipe(range(5)).shuffle(2)
    assert sorted(pipe) == [0, 1, 2, 3, 4]


def test_limit():
    pipe = Pipe(range(5)).limit(3)
    assert list(pipe) == [0, 1, 2]


def test_read_lines():
    import os
    import tempfile

    with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
        f.write("1\n2\n3\n4\n5")
        f.close()
        pipe = Pipe([f.name]).read_lines()
        assert list(pipe) == ["1", "2", "3", "4", "5"]
        os.remove(f.name)


def test_write_lines():
    import os
    import tempfile

    with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
        pipe = Pipe(range(5)).write_lines(f.name)
        list(pipe)
        with open(f.name, "r") as f2:
            assert f2.read() == "0\n1\n2\n3\n4\n"
        os.remove(f.name)


def test_filter_cached():
    import os
    import tempfile

    with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
        pipe = Pipe(range(5)).filter_cached(f.name)
        assert list(pipe) == [0, 1, 2, 3, 4]
        pipe = Pipe(range(6)).filter_cached(f.name)
        assert list(pipe) == [5]
        os.remove(f.name)


def test_read_csv():
    import os
    import tempfile

    with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
        f.write("1,2,3,4,5\n6,7,8,9,10")
        f.close()
        pipe = Pipe([f.name]).read_csv()
        assert list(pipe) == [["1", "2", "3", "4", "5"], ["6", "7", "8", "9", "10"]]
        os.remove(f.name)


def test_write_csv():
    import os
    import tempfile

    with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
        pipe = Pipe([[1, 2, 3, 4, 5], [6, 7, 8, 9, 10]]).write_csv(f.name)
        list(pipe)
        with open(f.name, "r") as f2:
            assert f2.read() == "1,2,3,4,5\n6,7,8,9,10\n"
        os.remove(f.name)

    # Test with dict
    with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
        pipe = Pipe(
            [
                {"a": 1, "b": 2, "c": 3, "d": 4, "e": 5},
                {"a": 6, "b": 7, "c": 8, "d": 9, "e": 10},
            ]
        ).write_csv(f.name)
        list(pipe)
        with open(f.name, "r") as f2:
            assert f2.read() == "a,b,c,d,e\n1,2,3,4,5\n6,7,8,9,10\n"
        os.remove(f.name)


def test_read_files():
    import os
    import tempfile

    with tempfile.TemporaryDirectory() as d:
        with open(os.path.join(d, "f1.txt"), "w") as f1:
            f1.write("a")
        with open(os.path.join(d, "f2.txt"), "w") as f2:
            f2.write("b")
        pipe = Pipe([f"{d}/*.txt"]).read_files()
        assert list(pipe) == [f1.name, f2.name]
        os.remove(f1.name)
        os.remove(f2.name)

    # Test cache
    with tempfile.TemporaryDirectory() as d:
        with open(os.path.join(d, "f1.txt"), "w") as f1:
            f1.write("a")
        with open(os.path.join(d, "f2.txt"), "w") as f2:
            f2.write("b")
        assert not os.path.exists(f"{d}/cache.txt")
        pipe = Pipe([f"{d}/*.txt"]).read_files(cache_filepath=f"{d}/cache.txt")
        assert list(pipe) == [f1.name, f2.name]
        # Check content of cache
        with open(f"{d}/cache.txt", "r") as f:
            assert f.read() == f"{f1.name}\n{f2.name}\n"
        pipe = Pipe([f"{d}/*.txt"]).read_files(cache_filepath=f"{d}/cache.txt")
        assert list(pipe) == [f1.name, f2.name]
        os.remove(f"{d}/cache.txt")

    # Test shuffle
    with tempfile.TemporaryDirectory() as d:
        with open(os.path.join(d, "f1.txt"), "w") as f1:
            f1.write("a")
        with open(os.path.join(d, "f2.txt"), "w") as f2:
            f2.write("b")
        pipe = Pipe([f"{d}/*.txt"]).read_files(shuffle=True)
        assert sorted(list(pipe)) == sorted([f1.name, f2.name])
        os.remove(f1.name)
        os.remove(f2.name)



import os
from typing import Callable, Iterable, Iterator, Optional, TypeVar

from pipd import Function, Pipe

from .read_lines import read_lines

T = TypeVar("T")


class FilterCached(Function):
    def __init__(self, filepath: str, key: Optional[Callable] = None) -> None:
        self.filepath = filepath
        self.key = key

    def __call__(self, items: Iterable[T]) -> Iterator[T]:
        cache = (
            set(read_lines(self.filepath)) if os.path.exists(self.filepath) else set()
        )
        with open(self.filepath, "a") as file:
            for item in items:
                value = str(self.key(item) if self.key is not None else item)
                if value not in cache:
                    cache.add(value)
                    file.write(value + "\n")
                    yield item


Pipe.add_fn(FilterCached)



from __future__ import annotations

import random
import re
import traceback
from typing import Any, Callable, Dict, Iterator, Optional, Sequence, Type


def is_iterable(obj):
    try:
        iter(obj)
        return True
    except TypeError:
        return False


def log_traceback_and_continue(exception: Exception):
    message = "Exception in Pipe, logging traceback continuing:\n"
    message += "".join(
        traceback.format_exception(type(exception), exception, exception.__traceback__)
    )
    print(message)


def camelcase_to_snakecase(name):
    return re.sub(r"([a-z])([A-Z])", r"\1_\2", name).lower()


def identity(x: Any) -> Any:
    return x


def compose(*funcs: Callable[[Any], Any]) -> Callable[[Any], Any]:
    def composed(arg: Any) -> Any:
        for func in funcs:
            arg = func(arg)
        return arg

    return composed


def merge_rand(
    *iterators: Iterator[Any], weights: Optional[Sequence[float]] = None
) -> Iterator[Any]:
    iterator_list = list(iterators)
    weights = [1] * len(iterator_list) if weights is None else weights
    while iterator_list:
        next_iter: Iterator = random.choices(iterator_list, weights=weights)[0]
        try:
            yield next(next_iter)
        except StopIteration:
            break


class Function:
    def __call__(self, items) -> Iterator:
        raise NotImplementedError


class PipeMeta(type):
    def __getattr__(cls, name):
        def method(*args, **kwargs):
            return cls().__getattr__(name)(*args, **kwargs)

        return method


class Pipe(metaclass=PipeMeta):
    functions: Dict[str, Type[Function]] = {}

    def __init__(
        self,
        *iterable,
        function: Optional[Callable] = identity,
        handler: Optional[Callable] = log_traceback_and_continue,
    ):
        if len(iterable) == 1 and is_iterable(iterable[0]):
            iterable = iterable[0]
        self.iterable = iterable
        self.function = function
        self.handler = handler

    def __iter__(self):
        for item in self.function(iter(self.iterable)):
            try:
                yield item
            except Exception as e:
                self.handler(e)

    def close(self):
        # Necessary to use `yield from` on a Pipe object
        pass

    def list(self):
        return list(self)

    def __getattr__(self, name):
        def method(*args, **kwargs):
            cls = self.__class__
            function = self.functions[name](*args, **kwargs)
            return cls(self.iterable, function=compose(self.function, function))

        return method

    def __call__(self, *iterable):
        return self.__class__(*iterable, function=self.function)

    @classmethod
    def add_fn(cls, fn: Type[Function], name: Optional[str] = None):
        fn_name = camelcase_to_snakecase(fn.__name__) if name is None else name
        cls.functions[fn_name] = fn
